@@ -1,42 +1,28 @@
 /*
- * Copyright (c) 2000 Silicon Graphics, Inc.  All Rights Reserved.
- * 
- * This program is free software; you can redistribute it and/or modify it
- * under the terms of version 2 of the GNU General Public License as
+ * Copyright (c) 2000-2001 Silicon Graphics, Inc.
+ * All Rights Reserved.
+ *
+ * This program is free software; you can redistribute it and/or
+ * modify it under the terms of the GNU General Public License as
  * published by the Free Software Foundation.
- * 
- * This program is distributed in the hope that it would be useful, but
- * WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
- * 
- * Further, this software is distributed without any warranty that it is
- * free of the rightful claim of any third person regarding infringement
- * or the like.  Any license provided herein, whether implied or
- * otherwise, applies only to this software file.  Patent licenses, if
- * any, provided herein do not apply to combinations of this program with
- * other software, or any other product whatsoever.
- * 
- * You should have received a copy of the GNU General Public License along
- * with this program; if not, write the Free Software Foundation, Inc., 59
- * Temple Place - Suite 330, Boston MA 02111-1307, USA.
- * 
- * Contact information: Silicon Graphics, Inc., 1600 Amphitheatre Pkwy,
- * Mountain View, CA  94043, or:
- * 
- * http://www.sgi.com 
- * 
- * For further information regarding this notice, see: 
- * 
- * http://oss.sgi.com/projects/GenInfo/SGIGPLNoticeExplan/
+ *
+ * This program is distributed in the hope that it would be useful,
+ * but WITHOUT ANY WARRANTY; without even the implied warranty of
+ * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+ * GNU General Public License for more details.
+ *
+ * You should have received a copy of the GNU General Public License
+ * along with this program; if not, write the Free Software Foundation,
+ * Inc.,  51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
  */
 
-#include <libxfs.h>
-#include <jdm.h>
+#include <xfs/xfs.h>
+#include <xfs/jdm.h>
 
 #include <sys/mman.h>
 #include <errno.h>
 #include <memory.h>
-#include <limits.h> 
+#include <limits.h>
 
 #include "types.h"
 #include "mlog.h"
@@ -44,6 +30,7 @@
 #include "node.h"
 #include "mmap.h"
 
+#define max( a, b )	( ( ( a ) > ( b ) ) ? ( a ) : ( b ) )
 #define min( a, b )	( ( ( a ) < ( b ) ) ? ( a ) : ( b ) )
 
 extern size_t pgsz;
@@ -158,6 +145,9 @@ struct node_hdr {
 		/* the backing store is partitoned into segment, which
 		 * can be mapped into VM windows  by the win abstraction
 		 */
+	size64_t nh_segtblsz;
+		/* the estimated size of the entire segment table.
+		 */
 	size_t nh_winmapmax;
 		/* maximum number of windows which can be mapped
 		 */
@@ -201,18 +191,16 @@ node_init( intgen_t fd,
 	   ix_t nodehkix,
 	   size_t nodealignsz,
 	   size64_t vmsz,
-	   size64_t dirs_nondirs_cnt,
-	   bool_t largewindowpr )
+	   size64_t dirs_nondirs_cnt )
 {
 	size64_t nodesz;
+	size64_t winmap_mem;
 	size64_t segsz;
+	size64_t segtablesz;
 	size64_t nodesperseg;
-	size64_t init_nodesperseg;
-	size64_t multiple;
+	size64_t minsegsz;
 	size64_t winmapmax;
-	unsigned long i;
-  	intgen_t rval;
- 	bool_t reached_min; /* reached minimum limit */
+	intgen_t rval;
 
 	/* sanity checks
 	 */
@@ -228,6 +216,41 @@ node_init( intgen_t fd,
 		 * link it in the free list.
 		 */
 
+	/* adjust the user's node size to meet user's alignment constraint
+	*/
+	nodesz = ( usrnodesz + nodealignsz - 1 ) & ~( nodealignsz - 1 );
+
+#define	WINMAP_MAX	20	/* maximum number of windows to use */
+#define	WINMAP_MIN	4	/* minimum number of windows to use */
+#define	HARDLINK_FUDGE	1.2	/* approx 1.2 hard links per file */
+
+	/* Calculate the expected size of the segment table using the number
+	 * of dirs and non-dirs.  Since we don't know how many hard-links
+	 * there will be, scale the size upward using HARDLINK_FUDGE.
+	 */
+
+	segtablesz = ( (size64_t)(HARDLINK_FUDGE * (double)dirs_nondirs_cnt) * nodesz);
+
+	/* Figure out how much memory is available for use by winmaps, and
+	 * use that to pick an appropriate winmapmax, segsz, and nodesperseg,
+	 * the goal being that if at all possible we want the entire segment
+	 * table to be mapped so that we aren't constantly mapping and
+	 * unmapping winmaps.  There must be at least WINMAP_MIN winmaps
+	 * because references can be held on more than one winmap at the
+	 * same time.  More winmaps are generally better to reduce the
+	 * number of nodes that are unmapped if unmapping does occur.
+	 */
+
+	minsegsz = pgsz * nodesz;	/* must be pgsz and nodesz multiple */
+	winmap_mem = min(vmsz, segtablesz);
+	segsz = (((winmap_mem / WINMAP_MAX) + minsegsz - 1) / minsegsz) * minsegsz;
+	segsz = max(segsz, minsegsz);
+
+	nodesperseg = segsz / nodesz;
+
+	winmapmax = min(WINMAP_MAX, vmsz / segsz);
+	winmapmax = max(winmapmax, WINMAP_MIN);
+
 	/* map the abstraction header
 	 */
 	ASSERT( ( NODE_HDRSZ & pgmask ) == 0 );
@@ -238,131 +261,20 @@ node_init( intgen_t fd,
 					    NODE_HDRSZ,
 					    fd,
 					    off );
-
 	if ( node_hdrp == (node_hdr_t *)-1 ) {
-	    mlog( MLOG_NORMAL | MLOG_ERROR,
-		  "unable to map node hdr of size %d: %s\n",
+	    mlog( MLOG_NORMAL | MLOG_ERROR, _(
+		  "unable to map node hdr of size %d: %s\n"),
 		  NODE_HDRSZ,
 		  strerror( errno ));
 	    return BOOL_FALSE;
 	}
 
-	/* adjust the user's node size to meet user's alignment constraint
-	 */
-	nodesz = ( usrnodesz + nodealignsz - 1 ) & ~( nodealignsz - 1 );
-
-
-	if ( vmsz > ( size64_t )SIZEMAX ) {
-		vmsz = ( size64_t )SIZEMAX; /* be reasonable! */
-	}
-
-	/* Calculate the segment size using the number of
-         * dirs and nondirs if largemap option else using a
-         * hard coded number.
-	 * The segment size must be an integral multiple of pgsz.
-	 */
-	if (largewindowpr) {
-	    /* Scale the dirs_nondirs_cnt because the number of entries
-             * also include hard links, of which I don't know how
-	     * many there are.
-	     */
-	    mlog( MLOG_DEBUG, "attempt mmaping entire tree\n");
-	    nodesperseg = (size64_t) (1.2 * (double)dirs_nondirs_cnt);
-	}
-	else {
-	    nodesperseg = NODESPERSEGMIN;
-	}
-	init_nodesperseg = nodesperseg;
-
-        /* 
-	 * Find largest fitting segsz.
-         * Loop trying init_nodesperseg/i for i=2,3,4,...
-	 * until mmap() is not failing with ENOMEM
-	 * or we have gone down to NODESPERSEGMIN
-	 * in which case, then giveup with error.
-         *
-         */
-	reached_min=BOOL_FALSE;
-	for (i=2;;i++) {
-	    char *map;
-
-	    if (nodesperseg <= NODESPERSEGMIN) {
-		nodesperseg = NODESPERSEGMIN;
-		reached_min = BOOL_TRUE;
-	    }
-	    segsz = nodesz * nodesperseg;
-	    multiple = pgsz * nodesz;
-	    /* ensure a multiple of pgsz and nodesz */
-	    segsz = (segsz / multiple + 1) * multiple;
-	    nodesperseg = segsz / nodesz;
-	    ASSERT( ! ( segsz % pgsz ));
-
-	    /* calculate the maximum number of windows which may be mapped.
-	     * base on vmsz, which is this abstraction's share of VM.
-	     */
-	    winmapmax = ( size_t )vmsz / segsz;
-	    ASSERT( winmapmax > 0 );
-
-	    /* segsz needs to fit into a size_t for mmap */
-	    if (segsz > SSIZE_MAX && !reached_min) {
-		goto try_smaller; 
-	    }
-
-	    /* try mem mapping sement of segsz 
-	     */
-	    map = (char *)mmap64( 0,
-				(size_t) segsz,
-				PROT_READ | PROT_WRITE,
-				MAP_SHARED,
-				fd,
-				off + NODE_HDRSZ);
-	    if ( map == (char *)-1 ) {
-		if ( (errno == ENOMEM||errno == ENXIO) && !reached_min ) {
-try_smaller:
-		    /* 
-		     * Try a smaller region as it failed and
-		     * we haven't reached the minimum limit.
-		     * Go for fitting into i number of maps.
-		     */
-		    nodesperseg = init_nodesperseg / i;
-		    mlog( MLOG_NORMAL | MLOG_NOTE,
-			  "unable to map tree segment of size %lld: "
-			  "retry with smaller size\n",
-			  segsz);
-		    continue;
-		}
-		else {
-		    mlog( MLOG_NORMAL | MLOG_ERROR,
-			  "unable to map tree segment of size %d: %s\n",
-			  segsz,
-			  strerror( errno ));
-		    return BOOL_FALSE;
-		}
-	    }
-	    else {
-		/* 
-		 * unmap the segment 
-		 * don't need it yet 
-		 * just trying to find largest fitting seg size
-		 */
-		if (munmap(map, segsz) == -1) {
-		    mlog( MLOG_NORMAL | MLOG_ERROR,
-			  "unable to unmap tree segment of size %lld: %s\n",
-			  segsz,
-			  strerror( errno ));
-		    return BOOL_FALSE;
-		}
-	    }
-	    break;
-	} /*for*/
-	ASSERT( ( size64_t )segsz <= OFF64MAX );
-			/* avoid sign extention questions */
-
 	/* initialize and save persistent context.
 	 */
 	node_hdrp->nh_nodesz = nodesz;
 	node_hdrp->nh_nodehkix = nodehkix;
 	node_hdrp->nh_segsz = segsz;
+	node_hdrp->nh_segtblsz = segtablesz;
 	node_hdrp->nh_winmapmax = winmapmax;
 	node_hdrp->nh_nodesperseg = nodesperseg;
 	node_hdrp->nh_nodealignsz = nodealignsz;
@@ -387,8 +299,8 @@ try_smaller:
 			    +
 			    ( off64_t )node_hdrp->nh_segsz );
 	if ( rval ) {
-		mlog( MLOG_NORMAL | MLOG_ERROR | MLOG_TREE,
-		      "unable to autogrow first node segment: %s (%d)\n",
+		mlog( MLOG_NORMAL | MLOG_ERROR | MLOG_TREE, _(
+		      "unable to autogrow first node segment: %s (%d)\n"),
 		      strerror( errno ),
 		      errno );
 		return BOOL_FALSE;
@@ -399,28 +311,24 @@ try_smaller:
 	win_init( fd,
 		  node_hdrp->nh_firstsegoff,
 		  segsz,
+		  segtablesz,
 		  winmapmax );
 	
 	/* announce the results
 	 */
 	mlog( MLOG_DEBUG | MLOG_TREE,
 	      "node_init:"
-	      " usrnodesz = %u (0x%x)"
-	      " nodesz = %u (0x%x)"
+	      " vmsz = %llu (0x%llx)"
 	      " segsz = %u (0x%x)"
+	      " segtblsz = %llu (0x%llx)"
 	      " nodesperseg = %u (0x%x)"
-	      " winmapmax = %u (0x%x)"
+	      " winmapmax = %llu (0x%llx)"
 	      "\n",
-	      usrnodesz,
-	      usrnodesz,
-	      nodesz,
-	      nodesz,
-	      segsz,
-	      segsz,
-	      nodesperseg,
-	      nodesperseg,
-	      winmapmax,
-	      winmapmax );
+	      vmsz, vmsz,
+	      segsz, segsz,
+	      segtablesz, segtablesz,
+	      nodesperseg, nodesperseg,
+	      winmapmax, winmapmax );
 
 	return BOOL_TRUE;
 }
@@ -442,8 +350,8 @@ node_sync( intgen_t fd, off64_t off )
 					    fd,
 					    off );
 	if ( node_hdrp == (node_hdr_t *)-1 ) {
-		mlog( MLOG_NORMAL | MLOG_ERROR,
-		      "unable to map node hdr of size %d: %s\n",
+		mlog( MLOG_NORMAL | MLOG_ERROR, _(
+		      "unable to map node hdr of size %d: %s\n"),
 		      NODE_HDRSZ,
 		      strerror( errno ));
 		return BOOL_FALSE;
@@ -458,6 +366,7 @@ node_sync( intgen_t fd, off64_t off )
 	win_init( fd,
 		  node_hdrp->nh_firstsegoff,
 		  node_hdrp->nh_segsz,
+		  node_hdrp->nh_segtblsz,
 		  node_hdrp->nh_winmapmax );
 
 	return BOOL_TRUE;
@@ -570,9 +479,9 @@ node_alloc( void )
 					    +
 					    ( off64_t )node_hdrp->nh_segsz );
 			if ( rval ) {
-				mlog( MLOG_NORMAL | MLOG_WARNING | MLOG_TREE,
+				mlog( MLOG_NORMAL | MLOG_WARNING | MLOG_TREE, _(
 				      "unable to autogrow node segment %llu: "
-				      "%s (%d)\n",
+				      "%s (%d)\n"),
 				      node_hdrp->nh_virgsegreloff
 				      /
 				      ( off64_t )node_hdrp->nh_segsz,
@@ -581,15 +490,15 @@ node_alloc( void )
 			}
 		}
 	}
-	
+
 	/* map in window containing node at top of free list,
 	 * and adjust free list.
 	 */
 	nix = node_hdrp->nh_freenix;
 #ifdef TREE_DEBUG
 	mlog(MLOG_DEBUG | MLOG_TREE,
-	     "node_alloc(): win_map(%llu) and get head from node freelist\n",
-	     NIX2OFF(nix));
+	   "node_alloc(): win_map(%llu) and get head from node freelist\n",
+           NIX2OFF(nix));
 #endif
 	win_map( NIX2OFF( nix ), ( void ** )&p );
 	if (p == NULL)
